"""
Poor man's Artificial Neural Networks with minimized input.

ANNs that include the state as input variable together with an additional
variable, say `z`. The controller output is then the value `z*` for which
the ANN takes on the smallest value (under the current state). In other
words, the ANN is supposed to model the system's objective function. The
idea is similar to :mod:`~moptipyapps.dynamic_control.controllers.ann`, but
instead of using the output of the ANNs as controller values, we use the value
`z*` for which the output of the ANN becomes minimal as controller value.
"""

from typing import Final, Iterable

import numba  # type: ignore
import numpy as np
from numpy import inf

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System

#: the golden ratio
PHI: Final[float] = 0.5 * (np.sqrt(5.0) + 1.0)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __min_ann_3d_1o_1(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Minimize a poor man's ANN for 3d-spaces with one layer containing 1 node.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed. The ANN has one additional parameter besides the `state`.
    We then find the value of this parameter for which the ANN takes on the
    smallest value via an iterated bracket - golden ratio search.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    hl_1_1_in: Final[float] = (state * params[0:3]).sum()
    x_3: Final[float] = params[3]

# first estimate of the best value
    x_best: float = 0.0
    f_best: float = np.arctan(hl_1_1_in + x_3 * x_best)

# now do bracketing
    x_low = x_a = -1000.0
    f_a: float = np.arctan(hl_1_1_in + x_3 * x_a)
    if f_a < f_best:
        x_best = x_a
        f_best = f_a

    x_c = x_b = -990.0
    f_c = f_b = np.arctan(hl_1_1_in + x_3 * x_b)
    if f_b < f_best:
        x_best = x_b
        f_best = f_b

# we try optimization in every single bracket we find
    found_bracket: bool = True
    ever_found_backet: bool = False
    while found_bracket:

        found_bracket = False
        while x_b < 1000.0:
            x_c = x_b + 10.0
            f_c = np.arctan(hl_1_1_in + x_3 * x_c)
            if f_c < f_best:
                x_best = x_c
                f_best = f_c
            if (f_c > f_b) and (f_b < f_a):
                ever_found_backet = found_bracket = True
                break
            x_a = x_b
            f_a = f_b
            x_b = x_c
            f_b = f_c

    # use huge range -1e3 ... 1e3
        if found_bracket:
            x_low = np.nextafter(x_a, inf)
            x_high = np.nextafter(x_c, -inf)
        elif ever_found_backet:
            break
        else:
            x_high = x_c

    # golden ratio algorithm
        delta = x_high - x_low
        while delta > 1e-12:
            delta /= PHI
            x_cc = x_high - delta

            f_cc: float = np.arctan(hl_1_1_in + x_3 * x_cc)
            if f_cc < f_best:
                x_best = x_cc
                f_best = f_cc

            x_dd = x_low + delta
            f_dd: float = np.arctan(hl_1_1_in + x_3 * x_dd)
            if f_dd < f_best:
                x_best = x_dd
                f_best = f_dd

            if f_cc < f_dd:
                x_high = np.nextafter(x_dd, -inf)
            else:
                x_low = np.nextafter(x_cc, inf)
            delta = x_high - x_low

        # move forward to next bracket
        x_a = x_b
        f_a = f_b
        x_b = x_c
        f_b = f_c

    out[0] = x_best


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __min_ann_2d_1o_1(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Minimize a poor man's ANN for 2d-spaces with one layer containing 1 node.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed. The ANN has one additional parameter besides the `state`.
    We then find the value of this parameter for which the ANN takes on the
    smallest value via an iterated bracket - golden ratio search.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    hl_1_1_in: Final[float] = (state * params[0:2]).sum()
    x_2: Final[float] = params[2]

# first estimate of the best value
    x_best: float = 0.0
    f_best: float = np.arctan(hl_1_1_in + x_2 * x_best)

# now do bracketing
    x_low = x_a = -1000.0
    f_a: float = np.arctan(hl_1_1_in + x_2 * x_a)
    if f_a < f_best:
        x_best = x_a
        f_best = f_a

    x_c = x_b = -990.0
    f_c = f_b = np.arctan(hl_1_1_in + x_2 * x_b)
    if f_b < f_best:
        x_best = x_b
        f_best = f_b

# we try optimization in every single bracket we find
    found_bracket: bool = True
    ever_found_backet: bool = False
    while found_bracket:

        found_bracket = False
        while x_b < 1000.0:
            x_c = x_b + 10.0
            f_c = np.arctan(hl_1_1_in + x_2 * x_c)
            if f_c < f_best:
                x_best = x_c
                f_best = f_c
            if (f_c > f_b) and (f_b < f_a):
                ever_found_backet = found_bracket = True
                break
            x_a = x_b
            f_a = f_b
            x_b = x_c
            f_b = f_c

    # use huge range -1e3 ... 1e3
        if found_bracket:
            x_low = np.nextafter(x_a, inf)
            x_high = np.nextafter(x_c, -inf)
        elif ever_found_backet:
            break
        else:
            x_high = x_c

    # golden ratio algorithm
        delta = x_high - x_low
        while delta > 1e-12:
            delta /= PHI
            x_cc = x_high - delta

            f_cc: float = np.arctan(hl_1_1_in + x_2 * x_cc)
            if f_cc < f_best:
                x_best = x_cc
                f_best = f_cc

            x_dd = x_low + delta
            f_dd: float = np.arctan(hl_1_1_in + x_2 * x_dd)
            if f_dd < f_best:
                x_best = x_dd
                f_best = f_dd

            if f_cc < f_dd:
                x_high = np.nextafter(x_dd, -inf)
            else:
                x_low = np.nextafter(x_cc, inf)
            delta = x_high - x_low

        # move forward to next bracket
        x_a = x_b
        f_a = f_b
        x_b = x_c
        f_b = f_c

    out[0] = x_best


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __min_ann_3d_1o_2(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Minimize a poor man's ANN for 3d-spaces with one layer containing 2 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed. The ANN has one additional parameter besides the `state`.
    We then find the value of this parameter for which the ANN takes on the
    smallest value via an iterated bracket - golden ratio search.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    hl_1_1_in: Final[float] = (state * params[0:3]).sum()
    x_3: Final[float] = params[3]
    x_4: Final[float] = params[4]
    hl_1_2_in: Final[float] = (state * params[5:8]).sum()
    x_8: Final[float] = params[8]
    x_9: Final[float] = params[9]
    x_10: Final[float] = params[10]
    x_11: Final[float] = params[11]

# first estimate of the best value
    x_best: float = 0.0
    hl_1_1: float = np.arctan(hl_1_1_in + x_3 * x_best + x_4)
    hl_1_2: float = np.arctan(hl_1_2_in + x_8 * x_best + x_9)
    f_best: float = hl_1_1 * x_10 + hl_1_2 * x_11

# now do bracketing
    x_low = x_a = -1000.0
    hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_a + x_4)
    hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_a + x_9)
    f_a: float = hl_1_1 * x_10 + hl_1_2 * x_11
    if f_a < f_best:
        x_best = x_a
        f_best = f_a

    x_c = x_b = -990.0
    hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_b + x_4)
    hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_b + x_9)
    f_c = f_b = hl_1_1 * x_10 + hl_1_2 * x_11
    if f_b < f_best:
        x_best = x_b
        f_best = f_b

# we try optimization in every single bracket we find
    found_bracket: bool = True
    ever_found_backet: bool = False
    while found_bracket:

        found_bracket = False
        while x_b < 1000.0:
            x_c = x_b + 10.0
            hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_c + x_4)
            hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_c + x_9)
            f_c = hl_1_1 * x_10 + hl_1_2 * x_11
            if f_c < f_best:
                x_best = x_c
                f_best = f_c
            if (f_c > f_b) and (f_b < f_a):
                ever_found_backet = found_bracket = True
                break
            x_a = x_b
            f_a = f_b
            x_b = x_c
            f_b = f_c

    # use huge range -1e3 ... 1e3
        if found_bracket:
            x_low = np.nextafter(x_a, inf)
            x_high = np.nextafter(x_c, -inf)
        elif ever_found_backet:
            break
        else:
            x_high = x_c

    # golden ratio algorithm
        delta = x_high - x_low
        while delta > 1e-12:
            delta /= PHI
            x_cc = x_high - delta

            hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_cc + x_4)
            hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_cc + x_9)
            f_cc: float = hl_1_1 * x_10 + hl_1_2 * x_11
            if f_cc < f_best:
                x_best = x_cc
                f_best = f_cc

            x_dd = x_low + delta
            hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_dd + x_4)
            hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_dd + x_9)
            f_dd: float = hl_1_1 * x_10 + hl_1_2 * x_11
            if f_dd < f_best:
                x_best = x_dd
                f_best = f_dd

            if f_cc < f_dd:
                x_high = np.nextafter(x_dd, -inf)
            else:
                x_low = np.nextafter(x_cc, inf)
            delta = x_high - x_low

        # move forward to next bracket
        x_a = x_b
        f_a = f_b
        x_b = x_c
        f_b = f_c

    out[0] = x_best


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __min_ann_2d_1o_2(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Minimize a poor man's ANN for 2d-spaces with one layer containing 2 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed. The ANN has one additional parameter besides the `state`.
    We then find the value of this parameter for which the ANN takes on the
    smallest value via an iterated bracket - golden ratio search.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    hl_1_1_in: Final[float] = (state * params[0:2]).sum()
    x_2: Final[float] = params[2]
    x_3: Final[float] = params[3]
    hl_1_2_in: Final[float] = (state * params[4:6]).sum()
    x_6: Final[float] = params[6]
    x_7: Final[float] = params[7]
    x_8: Final[float] = params[8]
    x_9: Final[float] = params[9]

# first estimate of the best value
    x_best: float = 0.0
    hl_1_1: float = np.arctan(hl_1_1_in + x_2 * x_best + x_3)
    hl_1_2: float = np.arctan(hl_1_2_in + x_6 * x_best + x_7)
    f_best: float = hl_1_1 * x_8 + hl_1_2 * x_9

# now do bracketing
    x_low = x_a = -1000.0
    hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_a + x_3)
    hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_a + x_7)
    f_a: float = hl_1_1 * x_8 + hl_1_2 * x_9
    if f_a < f_best:
        x_best = x_a
        f_best = f_a

    x_c = x_b = -990.0
    hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_b + x_3)
    hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_b + x_7)
    f_c = f_b = hl_1_1 * x_8 + hl_1_2 * x_9
    if f_b < f_best:
        x_best = x_b
        f_best = f_b

# we try optimization in every single bracket we find
    found_bracket: bool = True
    ever_found_backet: bool = False
    while found_bracket:

        found_bracket = False
        while x_b < 1000.0:
            x_c = x_b + 10.0
            hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_c + x_3)
            hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_c + x_7)
            f_c = hl_1_1 * x_8 + hl_1_2 * x_9
            if f_c < f_best:
                x_best = x_c
                f_best = f_c
            if (f_c > f_b) and (f_b < f_a):
                ever_found_backet = found_bracket = True
                break
            x_a = x_b
            f_a = f_b
            x_b = x_c
            f_b = f_c

    # use huge range -1e3 ... 1e3
        if found_bracket:
            x_low = np.nextafter(x_a, inf)
            x_high = np.nextafter(x_c, -inf)
        elif ever_found_backet:
            break
        else:
            x_high = x_c

    # golden ratio algorithm
        delta = x_high - x_low
        while delta > 1e-12:
            delta /= PHI
            x_cc = x_high - delta

            hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_cc + x_3)
            hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_cc + x_7)
            f_cc: float = hl_1_1 * x_8 + hl_1_2 * x_9
            if f_cc < f_best:
                x_best = x_cc
                f_best = f_cc

            x_dd = x_low + delta
            hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_dd + x_3)
            hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_dd + x_7)
            f_dd: float = hl_1_1 * x_8 + hl_1_2 * x_9
            if f_dd < f_best:
                x_best = x_dd
                f_best = f_dd

            if f_cc < f_dd:
                x_high = np.nextafter(x_dd, -inf)
            else:
                x_low = np.nextafter(x_cc, inf)
            delta = x_high - x_low

        # move forward to next bracket
        x_a = x_b
        f_a = f_b
        x_b = x_c
        f_b = f_c

    out[0] = x_best


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __min_ann_3d_1o_3(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Minimize a poor man's ANN for 3d-spaces with one layer containing 3 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed. The ANN has one additional parameter besides the `state`.
    We then find the value of this parameter for which the ANN takes on the
    smallest value via an iterated bracket - golden ratio search.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    hl_1_1_in: Final[float] = (state * params[0:3]).sum()
    x_3: Final[float] = params[3]
    x_4: Final[float] = params[4]
    hl_1_2_in: Final[float] = (state * params[5:8]).sum()
    x_8: Final[float] = params[8]
    x_9: Final[float] = params[9]
    hl_1_3_in: Final[float] = (state * params[10:13]).sum()
    x_13: Final[float] = params[13]
    x_14: Final[float] = params[14]
    x_15: Final[float] = params[15]
    x_16: Final[float] = params[16]
    x_17: Final[float] = params[17]

# first estimate of the best value
    x_best: float = 0.0
    hl_1_1: float = np.arctan(hl_1_1_in + x_3 * x_best + x_4)
    hl_1_2: float = np.arctan(hl_1_2_in + x_8 * x_best + x_9)
    hl_1_3: float = np.arctan(hl_1_3_in + x_13 * x_best + x_14)
    f_best: float = hl_1_1 * x_15 + hl_1_2 * x_16 + hl_1_3 * x_17

# now do bracketing
    x_low = x_a = -1000.0
    hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_a + x_4)
    hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_a + x_9)
    hl_1_3 = np.arctan(hl_1_3_in + x_13 * x_a + x_14)
    f_a = hl_1_1 * x_15 + hl_1_2 * x_16 + hl_1_3 * x_17
    if f_a < f_best:
        x_best = x_a
        f_best = f_a

    x_c = x_b = -990.0
    hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_b + x_4)
    hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_b + x_9)
    hl_1_3 = np.arctan(hl_1_3_in + x_13 * x_b + x_14)
    f_c = f_b = hl_1_1 * x_15 + hl_1_2 * x_16 + hl_1_3 * x_17
    if f_b < f_best:
        x_best = x_b
        f_best = f_b

# we try optimization in every single bracket we find
    found_bracket: bool = True
    ever_found_backet: bool = False
    while found_bracket:

        found_bracket = False
        while x_b < 1000.0:
            x_c = x_b + 10.0
            hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_c + x_4)
            hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_c + x_9)
            hl_1_3 = np.arctan(hl_1_3_in + x_13 * x_c + x_14)
            f_c = hl_1_1 * x_15 + hl_1_2 * x_16 + hl_1_3 * x_17
            if f_c < f_best:
                x_best = x_c
                f_best = f_c
            if (f_c > f_b) and (f_b < f_a):
                ever_found_backet = found_bracket = True
                break
            x_a = x_b
            f_a = f_b
            x_b = x_c
            f_b = f_c

    # use huge range -1e3 ... 1e3
        if found_bracket:
            x_low = np.nextafter(x_a, inf)
            x_high = np.nextafter(x_c, -inf)
        elif ever_found_backet:
            break
        else:
            x_high = x_c

    # golden ratio algorithm
        delta = x_high - x_low
        while delta > 1e-12:
            delta /= PHI
            x_cc = x_high - delta
            hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_cc + x_4)
            hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_cc + x_9)
            hl_1_3 = np.arctan(hl_1_3_in + x_13 * x_cc + x_14)
            f_cc = hl_1_1 * x_15 + hl_1_2 * x_16 + hl_1_3 * x_17
            if f_cc < f_best:
                x_best = x_cc
                f_best = f_cc

            x_dd = x_low + delta
            hl_1_1 = np.arctan(hl_1_1_in + x_3 * x_dd + x_4)
            hl_1_2 = np.arctan(hl_1_2_in + x_8 * x_dd + x_9)
            hl_1_3 = np.arctan(hl_1_3_in + x_13 * x_dd + x_14)
            f_dd = hl_1_1 * x_15 + hl_1_2 * x_16 + hl_1_3 * x_17
            if f_dd < f_best:
                x_best = x_dd
                f_best = f_dd

            if f_cc < f_dd:
                x_high = np.nextafter(x_dd, -inf)
            else:
                x_low = np.nextafter(x_cc, inf)
            delta = x_high - x_low

        # move forward to next bracket
        x_a = x_b
        f_a = f_b
        x_b = x_c
        f_b = f_c

    out[0] = x_best


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __min_ann_2d_1o_3(state: np.ndarray, _: float,
                      params: np.ndarray, out: np.ndarray) -> None:
    """
    Minimize a poor man's ANN for 2d-spaces with one layer containing 3 nodes.

    We use :func:`numpy.arctan` as activation function. The output layer
    is just a weighted sum of the hidden layer and no other transformation
    is performed. The ANN has one additional parameter besides the `state`.
    We then find the value of this parameter for which the ANN takes on the
    smallest value via an iterated bracket - golden ratio search.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    hl_1_1_in: Final[float] = (state * params[0:2]).sum()
    x_2: Final[float] = params[2]
    x_3: Final[float] = params[3]
    hl_1_2_in: Final[float] = (state * params[4:6]).sum()
    x_6: Final[float] = params[6]
    x_7: Final[float] = params[7]
    hl_1_3_in: Final[float] = (state * params[8:10]).sum()
    x_10: Final[float] = params[10]
    x_11: Final[float] = params[11]
    x_12: Final[float] = params[12]
    x_13: Final[float] = params[13]
    x_14: Final[float] = params[14]

# first estimate of the best value
    x_best: float = 0.0
    hl_1_1: float = np.arctan(hl_1_1_in + x_2 * x_best + x_3)
    hl_1_2: float = np.arctan(hl_1_2_in + x_6 * x_best + x_7)
    hl_1_3: float = np.arctan(hl_1_3_in + x_10 * x_best + x_11)
    f_best: float = hl_1_1 * x_12 + hl_1_2 * x_13 + hl_1_3 * x_14

# now do bracketing
    x_low = x_a = -1000.0
    hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_a + x_3)
    hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_a + x_7)
    hl_1_3 = np.arctan(hl_1_3_in + x_10 * x_a + x_11)
    f_a: float = hl_1_1 * x_12 + hl_1_2 * x_13 + hl_1_3 * x_14
    if f_a < f_best:
        x_best = x_a
        f_best = f_a

    x_c = x_b = -990.0

    hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_b + x_3)
    hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_b + x_7)
    hl_1_3 = np.arctan(hl_1_3_in + x_10 * x_b + x_11)
    f_c = f_b = hl_1_1 * x_12 + hl_1_2 * x_13 + hl_1_3 * x_14
    if f_b < f_best:
        x_best = x_b
        f_best = f_b

# we try optimization in every single bracket we find
    found_bracket: bool = True
    ever_found_backet: bool = False
    while found_bracket:

        found_bracket = False
        while x_b < 1000.0:
            x_c = x_b + 10.0
            hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_c + x_3)
            hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_c + x_7)
            hl_1_3 = np.arctan(hl_1_3_in + x_10 * x_c + x_11)
            f_c = hl_1_1 * x_12 + hl_1_2 * x_13 + hl_1_3 * x_14
            if f_c < f_best:
                x_best = x_c
                f_best = f_c
            if (f_c > f_b) and (f_b < f_a):
                ever_found_backet = found_bracket = True
                break
            x_a = x_b
            f_a = f_b
            x_b = x_c
            f_b = f_c

    # use huge range -1e3 ... 1e3
        if found_bracket:
            x_low = np.nextafter(x_a, inf)
            x_high = np.nextafter(x_c, -inf)
        elif ever_found_backet:
            break
        else:
            x_high = x_c

    # golden ratio algorithm
        delta = x_high - x_low
        while delta > 1e-12:
            delta /= PHI
            x_cc = x_high - delta
            hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_cc + x_3)
            hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_cc + x_7)
            hl_1_3 = np.arctan(hl_1_3_in + x_10 * x_cc + x_11)
            f_cc: float = hl_1_1 * x_12 + hl_1_2 * x_13 + hl_1_3 * x_14
            if f_cc < f_best:
                x_best = x_cc
                f_best = f_cc

            x_dd = x_low + delta
            hl_1_1 = np.arctan(hl_1_1_in + x_2 * x_dd + x_3)
            hl_1_2 = np.arctan(hl_1_2_in + x_6 * x_dd + x_7)
            hl_1_3 = np.arctan(hl_1_3_in + x_10 * x_dd + x_11)
            f_dd: float = hl_1_1 * x_12 + hl_1_2 * x_13 + hl_1_3 * x_14
            if f_dd < f_best:
                x_best = x_dd
                f_best = f_dd

            if f_cc < f_dd:
                x_high = np.nextafter(x_dd, -inf)
            else:
                x_low = np.nextafter(x_cc, inf)
            delta = x_high - x_low

        # move forward to next bracket
        x_a = x_b
        f_a = f_b
        x_b = x_c
        f_b = f_c

    out[0] = x_best


def min_anns(system: System) -> Iterable[Controller]:
    """
    Create poor man's ANNs for modeling the objective function.

    Based on the dimensionality of the state space, we generate a set of ANNs
    with different numbers of layers and neurons. The weights of the neurons
    can then directly be optimized by a numerical optimization algorithm.
    This is, of course, probably much less efficient that doing some proper
    learning like back-propagation. However, it allows us to easily plug the
    ANNs into the same optimization routines as other controllers.

    :param system: the equations object
    :return: the ANNs
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    if system.state_dims == 2:
        return (Controller("min_ann_1", 2, 1, 3, __min_ann_2d_1o_1),
                Controller("min_ann_2", 2, 1, 10, __min_ann_2d_1o_2),
                Controller("min_ann_3", 2, 1, 15, __min_ann_2d_1o_3))
    if system.state_dims == 3:
        return (Controller("min_ann_1", 3, 1, 4, __min_ann_3d_1o_1),
                Controller("min_ann_2", 3, 1, 12, __min_ann_3d_1o_2),
                Controller("min_ann_3", 3, 1, 18, __min_ann_3d_1o_3))
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
A set of tools for ordering objects in 1 dimension.

Let's assume that we have `n` objects and a distance metric that can compute
the distance between two objects. We do not know and also do not care about in
how many dimension the objects exist - we just have objects and a distance
metric.

Now we want to find a one-dimensional order of the objects that reflects their
original distance-based topology. For each object `a`, we want that its
closest neighbor in the order is also its actual closest neighbor according to
the distance metric. It's second-closest neighbor should be the actual
second-closest neighbor according to the distance metric. And so on.

Since we only care about the object order and do not want to metrically map
the distances to one dimension, we can represent the solution as permutation
of natural numbers.

Of course, in a one-dimensional order, each object has exactly two closest
neighbors (the one on its left and the one on its right) unless it is situated
either at the beginning or end of the order, in which case it has exactly one
closest neighbor. Based on the actual distance metric, an object may have any
number of closest neighbors, maybe only one, or maybe three equally-far away
objects. So it is not clear whether a perfect mapping to the one-dimensional
permutations even exists.

But we can try to find one that comes as close as possible to the real deal.

Another way to describe this problem is as follows:
Imagine that you have `n` objects and only know their mutual distances.
You want to arrange them on a one-dimensional axis in a way that does sort of
reflect their neighborhood structure in whatever space they are originally
located in.

The goal of solving this one-dimensional ordering problem is then to arrange
`n` objects on a 1-dimensional (e.g., horizontal) axis given a distance matrix
describing their location in a potentially high-dimensional or unstructured
space.
The objects should be arranged in such a way that, for each object,

- the nearest neighbors on the 1-dimensional axis are also the nearest
  neighbors in the original space (according to the distance matrix
  provided),
- the second nearest neighbors on the 1-dimensional axis are also the
  second nearest neighbors in the original space (according to the distance
  matrix provided),
- the third nearest neighbors on the 1-dimensional axis are also the third
  nearest neighbors in the original space (according to the distance matrix
  provided),
- and so on; with (quadratically) decreasing weights of neighbor distance
  ranks.

The original distances be limited to integers for the sake of simplicity, but
we may use floats as well if we want to. Either way, we do not care about the
actual precise distances (e.g., something like "0.001") between the objects
on either the one-dimensional nor the original space. Only about the distance
ranks, i.e., about "2nd nearest neighbor," but not "0.012 distance units
away." The solutions of this problem are thus permutations (orders) of the
objects. Of course, if we really want to plot the objects, such a permutation
can easily be translated to `x`-coordinates, say, by dividing the index of an
object by the number of objects, which nets values in `[0,1]`. But basically,
we reduce the task to finding permutations of objects that reflect the
neighbor structure of the original space as closely as possible.

If such a problem is solved correctly, then the arrangement on the
one-dimensional axis should properly reflect the arrangement of the objects in
the original space. Of course, solving this problem exactly may not actually
be possible, since an object on a one-dimensional axis may either have exactly
two `i`-nearest-neighbors (if it is at least `i` slots away from either end of
the permutation) or exactly `1` such neighbor, if it is closer that `i` units.
The object directly at the start of the permutation has only 1 nearest
neighbor (the object that comes next). That next object, however, has two,
namely the first object and the third object. In the original space where the
objects come from, however, there may be any number of "nearest neighbors."
Imagine a two-dimensional space where one object sits at the center of a
circle of other objects. Then all other objects are its nearest neighbors,
whereas an object on the circle either has exactly two nearest neighbors or,
maybe, in the odd situation that the radius equals a multiple of the
distance to the neighbors on the circle, three. Such a structure cannot be
represented exactly in one dimension.

But that's OK.
Because we mainly do this for visualization purposes anyway.
"""



"""
Synthesize some interesting starting points.

Here we have some basic functionality to synthesize starting points, i.e.,
training cases, for the dynamic systems control task.
The points synthesized here by function
:func:`make_interesting_starting_points` try to fulfill two goals:

1. the points should be as far away from each other as possible in the state
   space,
2. there should be points of many different distances from the state space
   origin, and
3. compared to the coordinates of the other points, the coordinates of the
   synthesized points should be sometimes smaller and sometimes larger (where
   sometimes should ideally mean "equally often").

These two goals are slightly contradicting and are achieved by forcing the
points to be located on rings of increasing distance from the origin via
:func:`interesting_point_transform` while maximizing their mean distance
to each other via :func:`interesting_point_objective`.
Since :func:`make_interesting_starting_points` is a bit slow, it makes sense
to pre-compute the points and then store them as array constants.
"""

from typing import Callable, Final, Iterable, cast

import numba  # type: ignore
import numpy as np
from moptipy.algorithms.so.vector.cmaes_lib import BiPopCMAES
from moptipy.api.execution import Execution
from moptipy.api.objective import Objective
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.console import logger


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def interesting_point_transform(
        x: np.ndarray, max_radius: float, dim: int) -> np.ndarray:
    """
    Transform interesting points.

    >>> xx = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12.0])
    >>> ppp = interesting_point_transform(xx, 10.0, 3)
    >>> print(ppp)
    [[0.6681531  1.33630621 2.00445931]
     [2.27921153 2.84901441 3.41881729]
     [3.76928033 4.30774895 4.84621757]
     [5.23423923 5.75766315 6.28108707]]
    >>> print(xx)
    [0.6681531  1.33630621 2.00445931 2.27921153 2.84901441 3.41881729
     3.76928033 4.30774895 4.84621757 5.23423923 5.75766315 6.28108707]
    >>> print([np.sqrt(np.square(pppp).sum()) for pppp in ppp])
    [2.5, 5.0, 7.5, 10.0]
    >>> ppp = interesting_point_transform(xx, 10.0, 3)
    >>> print(ppp)
    [[0.6681531  1.33630621 2.00445931]
     [2.27921153 2.84901441 3.41881729]
     [3.76928033 4.30774895 4.84621757]
     [5.23423923 5.75766315 6.28108707]]
    >>> print([np.sqrt(np.square(pppp).sum()) for pppp in ppp])
    [2.5, 5.0, 7.5, 10.0]
    >>> ppp = interesting_point_transform(xx, 10.0, 2)
    >>> print(ppp)
    [[0.74535599 1.49071198]
     [2.20132119 2.50305736]
     [3.200922   3.8411064 ]
     [4.39003072 5.01717796]
     [5.66154475 6.11484713]
     [6.75724629 7.3715414 ]]
    """
    n: Final[int] = len(x) // dim
    p: Final[np.ndarray] = np.reshape(x, (n, dim))
    for i in range(n):
        pp: np.ndarray = p[i, :]

        cur_radius: float = np.sqrt(np.square(pp).sum())
        if cur_radius <= 0.0:
            continue
        goal_radius: float = max_radius * ((i + 1) / n)
        if goal_radius != cur_radius:
            first: bool = True
            while True:
                old_radius = cur_radius
                pp2 = pp * goal_radius / cur_radius
                if np.array_equal(pp2, pp):
                    break
                cur_radius = np.sqrt(np.square(pp2).sum())
                if cur_radius == goal_radius:
                    pp = pp2
                    break
                if first:
                    first = False
                elif cur_radius > old_radius:
                    break
                pp = pp2
        p[i, :] = pp
    return p


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def interesting_point_objective(x: np.ndarray, other: np.ndarray,
                                max_radius: float, dim: int) -> float:
    """
    Compute the point diversity.

    1. the distance between the points
    2. the distance to the other points
    3. the coordinates in the different dimensions should be equally often
       smaller and larger than those of the starting points
    """
    pts: np.ndarray = interesting_point_transform(x, max_radius, dim)
    n: Final[int] = len(pts)
    f: float = 0.0
    score: int = 0
    for i in range(n):
        pt: np.ndarray = pts[i]
        scale: float = (((i + 1) / n) ** 2)  # we know this
        for j in range(i + 1, n):
            dst: float = np.sqrt(np.square(pt - pts[j]).sum())
            f += scale / dst if dst > 0.0 else 1e10
        for oth in other:
            dst = np.sqrt(np.square(pt - oth).sum())
            f += scale / dst if dst > 0.0 else 1e10
            for k, val in enumerate(pt):
                score += -1 if val < oth[k] else 1

    return f * np.log1p(abs(score))


def make_interesting_starting_points(
        n: int, other: Iterable[Iterable[float]],
        log: bool = True) -> np.ndarray:
    """
    Create some reasonably diverse starting points.

    :param n: the number of starting points
    :param other: the other points
    :param log: write log output
    :return: the starting points

    >>> p = make_interesting_starting_points(
    ...     3, np.array([[1.0, 2.0], [3.0, 2.9]]), False)
    >>> print(",".join(";".join(f"{x:.5f}" for x in row) for row in p))
    1.77767;0.33029,1.10482;3.44328,3.18515;-4.39064

    >>> p = make_interesting_starting_points(
    ...     3, np.array([[1.0, 2.0, 7.0], [3.0, 2.9, 1.1]]), False)
    >>> print(",".join(";".join(f"{x:.5f}" for x in row) for row in p))
    1.51392;-2.66567;0.86153,3.26154;-5.07757;2.03484,3.75627;5.88214;6.52310
    """
    other_points: Final[np.ndarray] = np.array(other)
    dim: Final[int] = other_points.shape[1]
    max_fes: Final[int] = 2048 + int(40 * (n ** (dim / 1.7)))

    max_radius: float = max(np.sqrt(np.square(pt).sum())
                            for pt in other_points)
    max_radius = max(max_radius ** 1.1, max_radius ** 0.6,
                     max_radius * 1.3)

    max_dim: float = np.max(np.abs(
        np.array(other_points).flatten()))
    max_dim = max(max_dim ** 1.2, max_dim ** (1.0 / 1.2),
                  max_dim ** 2.5, max_radius * 1.2)

    if log:
        logger(
            f"now determining {n} hopefully diverse samples of dimension "
            f"{dim} using {len(other_points)} other points, "
            f"max_radius={max_radius}, and max_dim={max_dim} for "
            f"{max_fes} FEs.")

    space: Final[VectorSpace] = VectorSpace(dim * n, -max_dim, max_dim)
    best: np.ndarray = space.create()

    class __Obj(Objective):
        def __init__(self):
            nonlocal other_points
            nonlocal max_radius
            nonlocal dim
            self.evaluate = cast(  # type: ignore
                Callable[[np.ndarray], float],
                lambda x, o=other_points, mr=max_radius, dd=dim:
                interesting_point_objective(x, o, mr, dd))

    with Execution().set_solution_space(space).set_rand_seed(1)\
            .set_algorithm(BiPopCMAES(space))\
            .set_max_fes(max_fes)\
            .set_objective(__Obj()).execute() as process:
        f = process.get_best_f()
        process.get_copy_of_best_x(best)
        best = interesting_point_transform(best, max_radius, dim)
        if log:
            logger(f"generated {n} points with objective {f}:\n{best!r}")
    return best



"""
A base class for implementing controllers.

A controller basically is a parameterizable function that receives the current
state and time of a :mod:`~moptipyapps.dynamic_control.system` as input and
computes one or multiple controller values as output. These controller values
are then used to influence how the state of the system changes in the next
iteration. In the dynamic systems control optimization task, the goal is to
find the right parameterization for the controller such that an
:mod:`~moptipyapps.dynamic_control.objective` is minimized.

Examples for different controllers for dynamic systems are given in package
:mod:`~moptipyapps.dynamic_control.controllers`.
"""


from typing import Callable, Final

import numpy as np
from moptipy.api.component import Component
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import check_to_int_range, type_error


class Controller(Component):
    """A class for governing a system via differential equations."""

    def __init__(self, name: str,
                 state_dims: int, control_dims: int, param_dims: int,
                 func: Callable[[np.ndarray, float, np.ndarray,
                                 np.ndarray], None] | None = None) -> None:
        """
        Initialize the system.

        :param name: the name of the system.
        :param state_dims: the state dimensions
        :param control_dims: the control dimensions
        :param param_dims: the parameter dimensions
        """
        super().__init__()
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        nn: Final[str] = sanitize_name(name)
        if nn != name:
            raise ValueError(
                f"sanitized name {nn!r} is different from name {name!r}.")
        #: the controller name
        self.name: Final[str] = name
        #: the dimensions of the state variable
        self.state_dims: Final[int] = check_to_int_range(
            state_dims, "state_dims", 2, 100)
        #: the dimensions of the controller output
        self.control_dims: Final[int] = check_to_int_range(
            control_dims, "control_dims", 1, 100)
        #: the dimensions of the controller parameter
        self.param_dims: Final[int] = check_to_int_range(
            param_dims, "param_dims", 1, 1_000)
        if func is not None:
            if not callable(func):
                raise type_error(func, "func", None, call=True)
            self.controller = func  # type: ignore

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("stateDims", self.state_dims)
        logger.key_value("controlDims", self.control_dims)
        logger.key_value("paramDims", self.param_dims)

    def __str__(self):
        """
        Get the name of this controller.

        :return: the name of this controller
        """
        return self.name

    def controller(self, state: np.ndarray,  # pylint: disable=E0202
                   time: float,  # pylint: disable=E0202
                   params: np.ndarray,  # pylint: disable=E0202
                   out: np.ndarray) -> None:  # pylint: disable=E0202
        """
        Compute the control value and store it in `out`.

        :param state: the state vector
        :param time: the time value
        :param params: the controller variables
        :param out: the output array to receive the controller values
        """

    def parameter_space(self) -> VectorSpace:
        """
        Create a vector space to represent the possible parameterizations.

        :return: a vector space for the possible parameterizations of this
            controller.
        """
        return VectorSpace(self.param_dims, -32.0, 32.0)



"""The setup and installation script."""

import re

from setuptools import setup

# We want to use our README.md as project description.
# However, we must fix all the references inside.
with open("README.md", encoding="utf-8-sig") as reader:
    old_lines = reader.readlines()

# It seems that the markdown parser does not auto-generate anchors. This means
# that we need to fix all references following the pattern `[xxx](#12-hello)`
# to `[xxx]({docu_url#hello)`, where `docu_url` is the url of our
# documentation. We do this with a regular expression `regex_search`.
new_lines = []
in_code: bool = False  # we only process non-code lines
# detects strings of the form [xyz](#123-bla) and gives \1=xyz and \2=bla
regex_search = re.compile("(\\[.+?])\\(#\\d+-(.+?)\\)")
regex_repl: str = \
    "\\1(https://thomasweise.github.io/moptipyapps/index.html#\\2)"
license_old: str = \
    "https://github.com/thomasWeise/moptipyapps/blob/main/LICENSE"
license_new: str = "https://thomasweise.github.io/moptipyapps/LICENSE.html"

for line in old_lines:
    line = line.rstrip()
    if in_code:
        if line.startswith("```"):
            in_code = False  # toggle to non-code
    else:
        if line.startswith("```"):
            in_code = True  # toggle to code
        else:  # fix all internal urls
            # replace links of the form "#12-bla" to "#bla"
            line = re.sub(regex_search, regex_repl, line)
            line = line.replace(license_old, license_new)
    new_lines.append(line)

# Now we can use the code in the setup.
setup(long_description="\n".join(new_lines),
      long_description_content_type="text/markdown")



"""
The `RobinX` example data for the Traveling Tournament Problem (TTP).

David Van Bulck of the Sports Scheduling Research group, part of the
Faculty of Economics and Business Administration at Ghent University, Belgium,
maintains "RobinX: An XML-driven Classification for Round-Robin Sports
Timetabling", a set of benchmark data instances and results of the TTP.
Here we include some of these TTP instances into our package.

This package does not offer anything useful except for holding the TTP
files. You can find the documentation and actual classes for solving and
playing around with the TSP in package :mod:`~moptipyapps.ttp`.

The original data of `robinX` can be found at
<https://robinxval.ugent.be/RobinX/travelRepo.php>
"""

from importlib import resources  # nosem
from typing import TextIO


def open_resource_stream(file_name: str) -> TextIO:
    """
    Open a TSPLib resource stream.

    :param file_name: the file name of the resource
    :return: the stream
    """
    return resources.open_text(package=f"{__package__}", resource=file_name)
